UNOCard = UNOCard or class()

function UNOCard:init(position, rotation)
	self._card_unit = World:spawn_unit(Idstring("units/pd2_mod_uno/playing_card"), Vector3())
	self._new_gui = World:newgui()
	self:add_workspace(self._card_unit:get_object(Idstring("c_box")))

	self._position = Vector3()
	self._position_target = Vector3()
	self._position_speed = 5
	self:set_position(position or Vector3())

	self._rotation = Rotation()
	self._rotation_target = Rotation()
	self._rotation_speed = 5
	self:set_rotation(rotation or Rotation())

	self._card_unit:unit_data().update = function(dg_self, unit, t, dt) self:update(t, dt) end
	self._card_unit:set_extension_update_enabled(Idstring("unit_data"), true)

	self._border_object = self._card_unit:get_object(Idstring("g_il"))

	self:set_border_enabled(false)

	self._owner = nil
end

function UNOCard:set_card_id(card_id)
	self._card_id = card_id

	self:set_tweak_data(tweak_data.uno.cards[card_id])
end

function UNOCard:set_tweak_data(tweak_data)
	self._tweak_data = clone(tweak_data)

	self:clear_gui()

	if not self._tweak_data then return end

	self:set_type(self._tweak_data.type, self._tweak_data.color)
	self:set_value(self._tweak_data.value)

	if self._tweak_data.text then
		self:set_inner_text(self._tweak_data.text)
		self:set_outer_text(self._tweak_data.text)
	end

	if self._tweak_data.texture then
		self:set_inner_texture(self._tweak_data.texture, self._tweak_data.texture_rect)
	end

	if self._tweak_data.outer_texture then
		self:set_outer_texture(self._tweak_data.outer_texture, self._tweak_data.outer_texture_rect)
	end

	if self._tweak_data.full_texture then
		self:set_full_texture(self._tweak_data.full_texture)
	end
end

function UNOCard:set_type(type_id, override_color)
	self._type = type_id

	local color = override_color or tweak_data.uno.colors[type_id] or Color("000000")
	self:set_background("units/pd2_mod_uno/uno_background", color)
end

function UNOCard:set_value(value)
	self._value = value

	if value then
		self._value = value

		self:set_inner_text(tostring(value))
		self:set_outer_text(tostring(value))
	end
end

function UNOCard:card_id()
	return self._card_id
end

function UNOCard:type()
	return self._type
end

function UNOCard:value()
	return self._value
end

function UNOCard:placement_rule()
	return self._tweak_data and self._tweak_data.placement_rule
end

function UNOCard:local_placement_clbk()
	return self._tweak_data and self._tweak_data.local_placement_clbk
end

function UNOCard:ai_placement_clbk()
	return self._tweak_data and self._tweak_data.ai_placement_clbk
end

function UNOCard:server_placement_clbk()
	return self._tweak_data and self._tweak_data.server_placement_clbk
end

function UNOCard:add_workspace(gui_object)
	local depth_offset = 0.02

	local x_size = 6.3
	local x_offset = x_size / 2
	local y_size = 9.8
	local y_offset = y_size / 2

	local gui_width = 200
	local gui_height = gui_width * (y_size/x_size)

	self._ws = self._new_gui:create_linked_workspace(
		gui_width, gui_height, gui_object,
		Vector3(x_offset, depth_offset, y_offset),
		Vector3(-x_size, 0, 0),
		Vector3(0, 0, -y_size)
	)
	self._gui = ExtendedPanel:new(self._ws:panel())
end

function UNOCard:clear_gui()
	if alive(self._gui) then
		self._gui:clear()
	end
end

function UNOCard:set_face_visible(state)
	if state then
		self._ws:show()
	else
		self._ws:hide()
	end
end

function UNOCard:set_background(texture, color)
	if self._background_bitmap and alive(self._background_bitmap) then
		self._gui:remove(self._background_bitmap)
	end

	local w = self._gui:w()
	local h = self._gui:h()

	self._background_bitmap = self._gui:bitmap({
		w = w,
		h = h,
		blend_mode = "normal",
		texture = texture,
		color = color
	})

	self._background_bitmap:set_center(w / 2, h / 2)
	self._background_bitmap:set_render_template(Idstring("VertexColorTextured"))
end

function UNOCard:set_inner_text(text)
	if self._inner_text and alive(self._inner_text) then
		self._gui:remove(self._inner_text)
	end

	local font = tweak_data.menu.pd2_massive_font
	local font_size = 120

	local w = self._gui:w()
	local h = self._gui:h()

	self._inner_text = self._gui:text({
		vertical = "center",
		align = "center",
		layer = 2,
		font = font,
		font_size = font_size,
		w = w,
		h = h,
		blend_mode = "normal",
		text = text,
		color = Color(1, 1, 1, 1)
	})

	self._inner_text:set_center(w / 2, h / 2)
	self._inner_text:set_render_template(Idstring("OverlayText"))
end

function UNOCard:set_outer_text(text)
	if self._outer_text_1 and alive(self._outer_text_1) then
		self._gui:remove(self._outer_text_1)
	end

	if self._outer_text_2 and alive(self._outer_text_2) then
		self._gui:remove(self._outer_text_2)
	end

	local margin = 50
	local font = tweak_data.menu.pd2_massive_font
	local font_size = 40

	local w = self._gui:w()
	local h = self._gui:h()

	self._outer_text_1 = self._gui:text({
		vertical = "top",
		align = "left",
		layer = 2,
		font = font,
		font_size = font_size,
		w = w - margin,
		h = h - margin,
		blend_mode = "normal",
		text = text,
		color = Color(1, 1, 1, 1)
	})

	self._outer_text_1:set_center(w / 2, h / 2)
	self._outer_text_1:set_render_template(Idstring("OverlayText"))

	self._outer_text_2 = self._gui:text({
		vertical = "top",
		align = "left",
		layer = 2,
		font = font,
		font_size = font_size,
		w = w - margin,
		h = h - margin,
		blend_mode = "normal",
		text = text,
		color = Color(1, 1, 1, 1)
	})

	self._outer_text_2:set_center(w / 2, h / 2)
	self._outer_text_2:set_rotation(180)
	self._outer_text_2:set_render_template(Idstring("OverlayText"))
end

function UNOCard:set_inner_texture(texture, texture_rect)
	local size = 120

	if self._inner_text and alive(self._inner_text) then
		self._gui:remove(self._inner_text)
	end

	if self._inner_texture and alive(self._inner_texture) then
		self._gui:remove(self._inner_texture)
	end

	self._inner_texture = self._gui:bitmap({
		layer = 2,
		h = size,
		w = size,
		blend_mode = "normal",
		texture = texture,
		texture_rect = texture_rect,
		color = Color(1, 1, 1, 1)
	})

	self._inner_texture:set_center(self._gui:w() / 2, self._gui:h() / 2)
	self._inner_texture:set_render_template(Idstring("VertexColorTextured"))
end

function UNOCard:set_outer_texture(texture, texture_rect)
	local margin = 50
	local size = 40
	local calc_margin = margin

	if self._outer_text_1 and alive(self._outer_text_1) then
		self._gui:remove(self._outer_text_1)
	end

	if self._outer_texture_1 and alive(self._outer_texture_1) then
		self._gui:remove(self._outer_texture_1)
	end

	self._outer_texture_1 = self._gui:bitmap({
		layer = 2,
		h = size,
		w = size,
		blend_mode = "normal",
		texture = texture,
		texture_rect = texture_rect,
		color = Color(1, 1, 1, 1)
	})

	self._outer_texture_1:set_center(calc_margin, calc_margin)
	self._outer_texture_1:set_render_template(Idstring("VertexColorTextured"))

	if self._outer_text_2 and alive(self._outer_text_2) then
		self._gui:remove(self._outer_text_2)
	end

	if self._outer_texture_2 and alive(self._outer_texture_2) then
		self._gui:remove(self._outer_texture_2)
	end

	self._outer_texture_2 = self._gui:bitmap({
		layer = 2,
		h = size,
		w = size,
		blend_mode = "normal",
		texture = texture,
		texture_rect = texture_rect,
		color = Color(1, 1, 1, 1)
	})

	self._outer_texture_2:set_center(self._gui:w() - calc_margin, self._gui:h() - calc_margin)
	self._outer_texture_2:set_render_template(Idstring("VertexColorTextured"))
end

function UNOCard:set_full_texture(texture)
	if self._full_texture and alive(self._full_texture) then
		self._gui:remove(self._full_texture)
	end

	local w = self._gui:w()
	local h = self._gui:h()

	self._full_texture = self._gui:bitmap({
		layer = 1,
		w = w,
		h = h,
		blend_mode = "normal",
		texture = texture,
		color = Color(1, 1, 1, 1)
	})

	self._full_texture:set_center(w / 2, h / 2)
	self._full_texture:set_render_template(Idstring("VertexColorTextured"))
end

function UNOCard:position()
	return self._position
end

function UNOCard:set_position(position)
	mvector3.set(self._position, position)
	mvector3.set(self._position_target, position)
	self._card_unit:set_position(position)
end

function UNOCard:rotation()
	return self._rotation
end

function UNOCard:set_rotation(rotation)
	mrotation.set_zero(self._rotation)
	mrotation.multiply(self._rotation, rotation)
	mrotation.set_zero(self._rotation_target)
	mrotation.multiply(self._rotation_target, rotation)
	self._card_unit:set_rotation(rotation)
end

function UNOCard:set_position_target(position, avoidance)
	mvector3.set(self._position_target, position)

	if avoidance then
		self._position_avoidance = avoidance
		self._position_raised = mvector3.copy(self._position)
	end

	self._reached_target = false
	self:evaluate_reached_target()
end

function UNOCard:set_rotation_target(rotation)
	mrotation.set_zero(self._rotation_target)
	mrotation.multiply(self._rotation_target, rotation)
	self._reached_target = false
	self:evaluate_reached_target()
end

function UNOCard:local_corners()
	local x_size = 7
	local y_size = 11

	local half_x = x_size/2
	local half_y = y_size/2

	local top_right = Vector3(half_x, 0, half_y):rotate_with(self._rotation)
	local top_left = Vector3(-half_x, 0, half_y):rotate_with(self._rotation)
	local bottom_left = Vector3(-half_x, 0, -half_y):rotate_with(self._rotation)
	local bottom_right = Vector3(half_x, 0, -half_y):rotate_with(self._rotation)

	return {
		top_right,
		top_left,
		bottom_left,
		bottom_right
	}
end

function UNOCard:corners(override_position)
	local corners = self:local_corners()
	for _, corner in pairs(corners) do
		mvector3.add(corner, override_position or self._position)
	end

	return corners
end

function UNOCard:set_owner(new_owner)
	if self._owner then
		self._owner:remove_card(self)
	end

	if new_owner then
		new_owner:add_card(self)
	end

	self._owner = new_owner
end

function UNOCard:evaluate_reached_target(do_owner_update)
	local distance = mvector3.distance(self._position, self._position_target)
	if not self._reached_target and distance < 0.5 then
		self._reached_target = true

		if do_owner_update and self._owner then
			self._owner:update_cards()
		end
	end

	if self._position_raised and distance < self._position_avoidance then
		mvector3.set(self._position, self._position_raised)
		self._position_avoidance = nil
		self._position_raised = nil
	end
end

function UNOCard:reached_target()
	return self._reached_target
end

function UNOCard:update(t, dt)
	self:evaluate_reached_target(true)

	mvector3.lerp(self._position, self._position, self._position_target, self._position_speed * dt)

	if self._position_raised then
		local lowest_z = self._position.z
		for _, corner in pairs(self:local_corners()) do
			if corner.z < lowest_z then
				lowest_z = corner.z
			end
		end

		mvector3.set(self._position_raised, self._position)
		mvector3.set_z(self._position_raised, self._position.z - lowest_z)
		self._card_unit:set_position(self._position_raised)
	else
		self._card_unit:set_position(self._position)
	end

	mrotation.slerp(self._rotation, self._rotation, self._rotation_target, self._rotation_speed * dt)
	self._card_unit:set_rotation(self._rotation)
end

function UNOCard:set_border_enabled(state)
	self._border_object:set_visibility(state)
end

function UNOCard:destroy()
	if self._card_unit and alive(self._card_unit) then
		World:delete_unit(self._card_unit)
	end

	if not alive(self._new_gui) then
		return
	end

	if alive(self._ws) then
		self._new_gui:destroy_workspace(self._ws)

		self._ws = nil
		self._gui = nil
	end

	self._new_gui = nil
end