UNOCardPile = UNOCardPile or class(UNOControllableCardHolder)

function UNOCardPile:init(position, rotation, face_down, orient_cards)
	UNOCardPile.super.init(self)

	self._position = position
	self._rotation = rotation

	self._face_down = face_down
	self._orient_cards = orient_cards
end

function UNOCardPile:update_cards()
	local offset = 0.024
	local rotation_offset = Rotation(0, 90, 0)
	if self._face_down then
		rotation_offset = Rotation(0, -90, 0)
	end

	local top_index = #self._cards - 4
	for index, card in pairs(self._cards) do
		card:set_position_target(self._position + (self._rotation:z() * index * offset), 0.1)

		local target_rotation = self._rotation * rotation_offset
		if not self._orient_cards then
			mrotation.set_yaw_pitch_roll(target_rotation,
				card:rotation():yaw(),
				target_rotation:pitch(),
				target_rotation:roll()
			)
		end

		card:set_rotation_target(target_rotation)
	end

	-- Lazy temp disable.
	for index, card in pairs(self._cards) do
		card:set_face_visible(not self._face_down and (index >= top_index))
		card:set_border_enabled(false)
	end
end

function UNOCardPile:distance(x, y)
	local card = self:top_card()
	local position = card and card:position() or self._position
	return self:_distance_helper(x, y, position)
end

function UNOCardPile:inside(x, y)
	local card = self:top_card()
	return card and self:inside_card(card, x, y)
end

function UNOCardPile:update(t, dt, x, y, selected)
	if self._controller and self._controller:is_turn() then
		local card = self:top_card()
		if not card then return end

		if selected then
			self:top_card():set_border_enabled(true)

			return "hand"
		else
			self:top_card():set_border_enabled(false)
		end
	else
		self:top_card():set_border_enabled(false)
	end
end

function UNOCardPile:mouse_pressed(button, x, y, selected)
	if self._controller and self._controller:is_turn() and button == Idstring("0") then
		if self:top_card() and selected then
			self:top_card():set_border_enabled(false)

			managers.uno_network:send_host_request("REQUEST_DRAW_CARD")
		end
	end
end