local husk_names = {
	wild = "units/pd2_dlc_wild/characters/npc_criminals_wild_1/player_criminal_wild_husk",
	joy = "units/pd2_dlc_joy/characters/npc_criminals_joy_1/player_criminal_joy_husk"
}

local function get_husk_unit(character_id)
	return husk_names[character_id] or tweak_data.blackmarket.characters[character_id].npc_unit:gsub("(.+)/npc_", "%1/player_") .. "_husk"
end

local function spawn_husk_unit(character_id, complete_outfit)
	local husk_unit = get_husk_unit(character_id)
	local husk_unit_idstring = Idstring(husk_unit)

	if not DB:has(Idstring("unit"), husk_unit_idstring) then return end

	local spawned_husk_unit = World:spawn_unit(husk_unit_idstring, Vector3())
	Network:detach_unit(spawned_husk_unit)

	local husk_base = spawned_husk_unit:base()
	local husk_movement = spawned_husk_unit:movement()
	local husk_inventory = spawned_husk_unit:inventory()
	local husk_sound = spawned_husk_unit:sound()
	local husk_damage = spawned_husk_unit:damage()

	husk_base.pre_destroy = function(self, unit)
		unit:movement():pre_destroy(unit)
		unit:inventory():pre_destroy(unit)
		UnitBase.pre_destroy(self, unit)
	end

	spawned_husk_unit:contour().add = function() end

	husk_movement.update = function(self, ...) end

	husk_movement._register_revive_SO = function() end
	husk_movement.set_need_assistance = function(self, need_assistance) self._need_assistance = need_assistance end
	husk_movement.set_need_revive = function(self, need_revive) self._need_revive = need_revive end

	husk_movement.update_visual_state = function (self)
		local character = complete_outfit.character

		local player_style_unit_name = tweak_data.blackmarket:get_player_style_value(complete_outfit.player_style, character, "third_unit")
		if player_style_unit_name then
			managers.dyn_resource:load(Idstring("unit"), Idstring(player_style_unit_name), DynamicResourceManager.DYN_RESOURCES_PACKAGE)
		end

		local gloves_unit_name = tweak_data.blackmarket:get_glove_value(complete_outfit.glove_id, character, "unit", complete_outfit.player_style, complete_outfit.suit_variation)
		if gloves_unit_name then
			managers.dyn_resource:load(Idstring("unit"), Idstring(gloves_unit_name), DynamicResourceManager.DYN_RESOURCES_PACKAGE)
		end

		complete_outfit.is_local_peer = false
		complete_outfit.mask_id = complete_outfit.mask.mask_id

		managers.criminals.set_character_visual_state(self._unit, character, complete_outfit)
	end

	husk_sound.say = function() end
	husk_sound.play = function() end
	husk_sound._play = function() end

	husk_damage:run_sequence_simple(managers.blackmarket:character_sequence_by_character_name(complete_outfit.character))

	husk_movement:set_character_anim_variables()
	husk_movement:update_visual_state()

	local level_data = managers.job and managers.job:current_level_data()
	if level_data and level_data.player_sequence then
		legs_damage:run_sequence_simple(level_data.player_sequence)
	end

	managers.groupai:state():unregister_criminal(spawned_husk_unit)

	return spawned_husk_unit
end

UNOCharacter = UNOCharacter or class()

function UNOCharacter:init(position, rotation, character_id, complete_outfit)
	self._husk_unit = spawn_husk_unit(character_id, complete_outfit)

	self:set_position(position)
	self:set_rotation(rotation)

	self._husk_movement = self._husk_unit:movement()
	self._husk_movement:play_state("civilian/spawn/loop")
	self._husk_movement:play_redirect("cm_pex_cuffed_sit_loop")
end

function UNOCharacter:position()
	return self._position
end

function UNOCharacter:set_position(position)
	self._position = position
	self._husk_unit:set_position(position)
end

function UNOCharacter:rotation()
	return self._rotation
end

function UNOCharacter:set_rotation(rotation)
	self._rotation = rotation
	self._husk_unit:set_rotation(rotation)
end