UNOController = UNOController or class()

function UNOController:init(id, position, rotation)
	self._local_id = id

	self._current_controllable = nil
	self._controllables = {}
	self._controllable_ids = {}

	self._position = position
	self._rotation = rotation

	self._cam_unit = World:spawn_unit(Idstring("units/gui/background_camera_01/access_camera"), self._position, self._rotation)
	self._cam_unit:camera():start()
	self._camera = self._cam_unit:camera()._camera
	self._shaker = self._cam_unit:camera()._shaker

	self._listener_id = managers.listener:add_listener("player_camera", self._camera, self._camera, nil, false)
	managers.listener:add_set("player_camera", { "player_camera" })
	self._listener_activation_id = managers.listener:activate_set("main", "player_camera")

	self._shaker:play("breathing", 0.2)

	self._fov_speed = 3
	self:set_fov(70)

	self._relative_mouse = Vector3(0.5, 0.5, 0)

	self._mouse_data = {
		mouse_press = callback(self, self, "mouse_pressed"),
		mouse_release = callback(self, self, "mouse_released"),
		id = "uno_controller"
	}

	managers.mouse_pointer:use_mouse(self._mouse_data)
end

function UNOController:is_turn()
	return managers.uno:is_turn(self._local_id)
end

function UNOController:get_controllable(controllable_id)
	return self._controllable_ids[controllable_id]
end

function UNOController:add_controllable(controllable, controllable_id)
	if controllable.setup_controller then
		controllable:setup_controller(self)
		table.insert(self._controllables, controllable)

		if controllable_id then
			self._controllable_ids[controllable_id] = controllable
		end
	end
end

function UNOController:remove_controllable(controllable)
	local new_controllables = {}
	for _, old_controllable in pairs(self._controllables) do
		if old_controllable ~= controllable then
			table.insert(new_controllables, old_controllable)
		end
	end
	self._controllables = new_controllables

	for index, old_controllable in pairs(self._controllable_ids) do
		if old_controllable == controllable then
			self._controllable_ids[index] = nil
		end
	end

	controllable:remove_controller()
end

function UNOController:set_fov(fov)
	self._fov = fov
	self._fov_target = fov
	self._camera:set_fov(fov)
end

function UNOController:set_fov_target(fov)
	self._fov_target = fov
end

function UNOController:get_relative_mouse()
	return self._relative_mouse
end

function UNOController:update(t, dt)
	self._fov = math.lerp(self._fov, self._fov_target, self._fov_speed * dt)
	self._camera:set_fov(self._fov)

	managers.mouse_pointer:_activate()

	local workspace = managers.mouse_pointer:workspace()
	local workspace_x, workspace_y = workspace:panel():size()
	local mouse_x, mouse_y = managers.mouse_pointer:world_position()

	mvector3.set_x(self._relative_mouse, mouse_x / workspace_x)
	mvector3.set_y(self._relative_mouse, mouse_y / workspace_y)

	local yaw_offset = (self._relative_mouse.x - 0.5) * 45
	local pitch_offset = (self._relative_mouse.y - 0.5) * 27

	self._cam_unit:set_rotation(self._rotation * Rotation(-yaw_offset, -pitch_offset, 0))

	self._current_controllable = nil
	local current_distance = 999999

	for _, controllable in pairs(self._controllables) do
		local controllable_distance = controllable:distance(mouse_x, mouse_y)

		if controllable_distance < current_distance and controllable:inside(mouse_x, mouse_y) then
			self._current_controllable = controllable
			current_distance = controllable_distance
		end
	end

	local cursor = nil
	for _, controllable in pairs(self._controllables) do
		cursor = controllable:update(t, dt, mouse_x, mouse_y, self._current_controllable == controllable) or cursor
	end

	managers.mouse_pointer:set_pointer_image(cursor or "arrow")
end

function UNOController:mouse_pressed(o, button, x, y)
	if button == Idstring("1") then
		self:set_fov_target(50)
	end

	for _, controllable in pairs(self._controllables) do
		controllable:mouse_pressed(button, x, y, self._current_controllable == controllable)
	end
end

function UNOController:mouse_released(o, button, x, y)
	if button == Idstring("1") then
		self:set_fov_target(70)
	end

	for _, controllable in pairs(self._controllables) do
		controllable:mouse_released(button, x, y, self._current_controllable == controllable)
	end
end

function UNOController:to_screen(position)
	return managers.mouse_pointer:workspace():world_to_screen(self._camera, position)
end

function UNOController:destroy()
	for _, controllable in pairs(self._controllables) do
		controllable:remove_controller()
	end

	if self._listener_id then
		managers.listener:remove_listener(self._listener_id)
		managers.listener:remove_set("player_camera")

		self._listener_id = nil
	end

	managers.mouse_pointer:remove_mouse("uno_controller")

	if self._cam_unit and alive(self._cam_unit) then
		self._cam_unit:camera():stop()
		World:delete_unit(self._cam_unit)
	end
end