UNOHand = UNOHand or class(UNOControllableCardHolder)

function UNOHand:init(position, rotation)
	UNOHand.super.init(self)

	self._current_card_index = nil
	self._current_card = nil

	self._position = position
	self._rotation = rotation
end

function UNOHand:_calculate_card_pos_rot(index, extra_back, extra_down, extra_left, cant_see_cards)
	local total_twist = 20

	local total_width = 40
	local total_height = 5

	local base_position = self._position + (self._rotation:y() * 40) + (self._rotation:z() * -20)
	local base_rotation = self._rotation * Rotation(180, 30, 0)

	if cant_see_cards then
		base_rotation = base_rotation * Rotation(0, -60, 0)
	end

	local card_count = #self._cards
	local percentage = index / (card_count + 1)
	local split_percentage = percentage - 0.5

	local card_twist = split_percentage * total_twist

	local card_back = (index * -0.05) + (extra_back or 0)
	local card_left = (split_percentage * total_width) - (extra_left or 0)
	local card_down = (math.sin(percentage * 180) * total_height) + (extra_down or 0)

	local new_pos = base_position + (base_rotation:y() * card_back) + (base_rotation:x() * -card_left) + (base_rotation:z() * card_down)
	local new_rot = base_rotation * Rotation(0, 0, -card_twist)

	return new_pos, new_rot
end

function UNOHand:update_cards(is_turn)
	local can_see_cards = not not self._controller

	for index, card in pairs(self._cards) do
		local extra_back = 0
		local extra_down = 0
		local extra_left = 0

		if self._current_card_index and index ~= self._current_card_index then
			extra_left = 3

			if index > self._current_card_index then
				extra_left = extra_left * -1
			end
		end

		if card == self._current_card then
			extra_back = 3
			extra_down = 2
			card:set_border_enabled(true)
		else
			card:set_border_enabled(false)
		end

		if not managers.uno:is_hand_turn(self) then
			extra_down = -8
		end

		local new_pos, new_rot = self:_calculate_card_pos_rot(index, extra_back, extra_down, extra_left, not can_see_cards)

		card:set_position_target(new_pos, 10)
		card:set_rotation_target(new_rot)

		card:set_face_visible(can_see_cards)
	end
end

function UNOHand:distance(x, y)
	local position = self._current_card and self._current_card:position() or self._position
	return self:_distance_helper(x, y, position)
end

function UNOHand:inside(x, y)
	for index, card in pairs(self._cards) do
		local card_pos, card_rot = self:_calculate_card_pos_rot(index, 0, 0, 0)
		if self:inside_card(card, x, y, card_pos) then
			return true
		end
	end

	return false
end

function UNOHand:update(t, dt, x, y, selected)
	if self._controller and self._controller:is_turn() then
		local new_card = nil
		local new_card_index = nil
		local current_distance = 999999
		for index, card in pairs(self._cards) do
			local card_pos, card_rot = self:_calculate_card_pos_rot(index, 0, 0)
			local card_distance = self:_distance_helper(x, y, card_pos)
			if card_distance < current_distance then
				new_card = card
				new_card_index = index
				current_distance = card_distance
			end

			card:set_border_enabled(false)
		end

		if self._current_card ~= new_card then
			self._current_card = new_card
			self._current_card_index = new_card_index
			self:update_cards()
		end

		if selected and self._current_card then
			self._current_card:set_border_enabled(true)

			return "hand"
		elseif self._current_card then
			self._current_card = nil
			self._current_card_index = nil
			self:update_cards()
		end
	else
		if self._current_card then
			self._current_card = nil
			self._current_card_index = nil
			self:update_cards()
		end

		for _, card in pairs(self._cards) do
			card:set_border_enabled(false)
		end
	end
end


function UNOHand:mouse_pressed(button, x, y, selected)
	if self._controller and self._controller:is_turn() and button == Idstring("0") then
		if self._current_card and selected then
			if managers.uno:can_play_card(self._current_card) then
				self._current_card:set_border_enabled(false)

				managers.uno:play_card_local(self._current_card, self._current_card_index)

				self._current_card = nil
				self._current_card_index = nil
			end
		end
	end
end