UNOTweakData = UNOTweakData or class()

function UNOTweakData:init()
	self:_init_events()
	self:_init_ai_parameters()
	self:_init_rules()
	self:_init_arenas()
	self:_init_cards()
end

function UNOTweakData:_init_events()
	self.intro_events = {
		{0.0, 3.0, "just_uno_intro"}
	}
	self.intro_card_cooldown = 0.2

	self.death_event = "just_uno_death"
	self.death_timer = 3
	self.post_death_timer = 2
end

function UNOTweakData:_init_ai_parameters()
	self.ai_parameters = {
		decision_time = 5,
		decision_time_random = 1
	}
end

function UNOTweakData:_init_rules()
	self.rules = {
		starting_hand_size = 7
	}
end

function UNOTweakData:_init_arenas()
	self.arenas = {}

	self.arenas.standard_table = {
		cameras = {
			{ Vector3(  5,  62, 130), Rotation(180, -37, 0) },
			{ Vector3( 54,  -7, 130), Rotation( 90, -37, 0) },
			{ Vector3( -5, -62, 130), Rotation(  0, -37, 0) },
			{ Vector3(-54,   7, 130), Rotation(-90, -37, 0) }
		},
		characters = {
			{ Vector3( -2,  40, -2), Rotation(180, 15, 0) },
			{ Vector3( 37,   0, -2), Rotation( 90, 15, 0) },
			{ Vector3( -2, -40, -2), Rotation(  0, 15, 0) },
			{ Vector3(-37,   0, -2), Rotation(-90, 15, 0) }
		},
		draw_pile = { Vector3(15, 15, 75.5), Rotation(45, 0, 0) },
		play_pile = { Vector3( 0,  0, 75.5), Rotation( 0, 0, 0) }
	}
end

function UNOTweakData:_init_cards()
	self:_init_placement_rules()
	self:_init_colors()

	self.cards = {}

	self:_init_special_cards()
	self:_init_number_cards()

	self:_init_card_obfuscation()

	self:_init_random_card_counts()
end

function UNOTweakData:_init_placement_rules()
	self.placement_rules = {}

	self.placement_rules.wildcard = function(bottom_type, bottom_value, top_type, top_value)
		return true
	end

	self.placement_rules.wildcard_value = function(bottom_type, bottom_value, top_type, top_value)
		return (bottom_type == top_type) or top_type == nil
	end

	self.placement_rules.wildcard_type = function(bottom_type, bottom_value, top_type, top_value)
		return (bottom_value == top_value) or top_value == nil
	end

	self.placement_rules.standard = function(bottom_type, bottom_value, top_type, top_value)
		return bottom_type == top_type or bottom_value == top_value or top_type == nil or top_value == nil
	end
end

function UNOTweakData:_init_colors()
	self.colors = {}

	self.colors.red = Color("D72600")
	self.colors.yellow = Color("ECD407")
	self.colors.blue = Color("0956BF")
	self.colors.green = Color("379711")
end

function UNOTweakData:_init_special_cards()
	local function local_color_change_clbk(card, card_index)
		local selector = managers.uno:local_selector()

		local index = 1
		for color_id, _ in pairs(tweak_data.uno.colors) do
			local color_card = managers.uno:spawn_card(card:position(), card:rotation())
			color_card:set_card_id(card:card_id())
			color_card:set_type(color_id)
			color_card:set_owner(selector)

			selector:set_callback(index, function(color_card)
				managers.uno_network:send_host_request("REQUEST_PLAY_CARD", {card:card_id(), card_index})
				managers.uno_network:send_host_request("REQUEST_CARD_COLOR", {color_card:type()})
				managers.uno_network:send_host_request("END_TURN")
			end)

			index = index + 1
		end

		return true
	end

	local function ai_color_change_clbk(player_id, card, card_index)
		managers.uno_network:send_peers_except_request(1, "SYNC_PLAY_CARD", {player_id, card:card_id(), card_index})

		local color_id = table.random_key(self.colors)
		managers.uno_network:send_peers_request("SYNC_CARD_COLOR", {color_id})
		managers.menu_component:post_event("loot_flip_card")

		return true
	end

	self.cards.color_change = {
		placement_rule = "wildcard_value",
		default_quantity = 4,
		local_placement_clbk = local_color_change_clbk,
		ai_placement_clbk = ai_color_change_clbk,
		full_texture = "units/pd2_mod_uno/uno_rgb_texture"
	}

	local function server_skip_turn()
		managers.uno:set_skip_count(1)
	end

	local function server_draw_count(count)
		managers.uno:set_draw_count(count)
	end

	local function server_draw_4()
		server_draw_count(4)
		server_skip_turn()
	end

	self.cards.wild_draw_4 = {
		placement_rule = "wildcard_value",
		default_quantity = 4,
		local_placement_clbk = local_color_change_clbk,
		ai_placement_clbk = ai_color_change_clbk,
		text = "+4",
		full_texture = "units/pd2_mod_uno/uno_rgb_texture",
		server_placement_clbk = server_draw_4,
	}

	local function generate_special_color_cards(special_id, special_info)
		for color, _ in pairs(self.colors) do
			self.cards[color .. "_" .. special_id] = table.merge({
				type = color,
				value = special_id,
				placement_rule = "standard",
				default_quantity = 1
			}, special_info)
		end
	end

	local function server_draw_2()
		server_draw_count(2)
		server_skip_turn()
	end

	generate_special_color_cards("draw_2", {
		text = "+2",
		server_placement_clbk = server_draw_2
	})

	generate_special_color_cards("skip", {
		server_placement_clbk = server_skip_turn,
		texture = "guis/textures/pd2/lock_infamy",
		outer_texture = "guis/textures/pd2/lock_infamy"
	})

	local function server_reverse()
		if managers.uno:player_count() > 2 then
			managers.uno:flip_turn_direction()
		else
			server_skip_turn()
		end
	end

	generate_special_color_cards("reverse", {
		server_placement_clbk = server_reverse,
		texture = "guis/textures/icon_loading",
		outer_texture = "guis/textures/icon_loading"
	})
end

function UNOTweakData:_init_number_cards()
	local function generate_number_card(value, type, quantity)
		self.cards[type .. "_" .. tostring(value)] = {
			type = type,
			value = value,
			placement_rule = "standard",
			default_quantity = quantity or 1,
		}
	end

	local function generate_number_cards(lower_bound, upper_bound, type, quantity)
		for index = lower_bound, upper_bound do
			generate_number_card(index, type, quantity)
		end
	end

	local function generate_standard_number_cards(type)
		generate_number_card(0, type, 1) -- Only One `0 Card`
		generate_number_cards(1, 9, type, 2) -- Two `1-9 Cards`
	end

	for color, _ in pairs(self.colors) do
		generate_standard_number_cards(color)
	end
end

-- Lazy way to hide what card is being sent.
function UNOTweakData:_init_card_obfuscation()
	self.to_obfuscated = {}
	self.from_obfuscated = {}

	for card_id, _ in pairs(self.cards) do
		local card_key = Idstring(card_id):key()

		self.to_obfuscated[card_id] = card_key
		self.from_obfuscated[card_key] = card_id
	end
end

function UNOTweakData:_init_random_card_counts()
	self._base_random_card_counts = {}
	self._base_total_cards = 0

	for card_id, card_data in pairs(self.cards) do
		local quantity = card_data.default_quantity or 0

		self._base_random_card_counts[card_id] = quantity
		self._base_total_cards = self._base_total_cards + quantity
	end

	self:_reset_random_card_counts()
end

function UNOTweakData:_reset_random_card_counts()
	self._live_random_card_counts = clone(self._base_random_card_counts)
	self._live_total_cards = self._base_total_cards
end

function UNOTweakData:get_arena(player_count)
	return self.arenas.standard_table
end

function UNOTweakData:get_random_card_id()
	local random_table = {}
	for card_id, card_count in pairs(self._live_random_card_counts) do
		for index = 1, card_count do
			table.insert(random_table, card_id)
		end
	end

	local final_key = table.random(random_table)

	self._live_random_card_counts[final_key] = self._live_random_card_counts[final_key] - 1
	self._live_total_cards = self._live_total_cards - 1

	if self._live_total_cards <= 0 then
		self:_reset_random_card_counts()
	end

	return final_key
end

Hooks:PostHook(TweakData, "_init_wip_tweak_data", "InitUNOTweakData", function(self)
	self.uno = UNOTweakData:new()
end)